"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var lodash_1 = require("lodash");
var pluralize = require("pluralize");
var TargetLanguage_1 = require("../TargetLanguage");
var Type_1 = require("../Type");
var Naming_1 = require("../Naming");
var Source_1 = require("../Source");
var Strings_1 = require("../Strings");
var ConvenienceRenderer_1 = require("../ConvenienceRenderer");
var RendererOptions_1 = require("../RendererOptions");
var unicode = require("unicode-properties");
var ObjectiveCTargetLanguage = /** @class */ (function (_super) {
    __extends(ObjectiveCTargetLanguage, _super);
    function ObjectiveCTargetLanguage() {
        var _this = _super.call(this, "Objective-C", ["objc", "objective-c", "objectivec"], "m") || this;
        _this._featuresOption = new RendererOptions_1.EnumOption("features", "Interface and implementation", [
            ["all", { interface: true, implementation: true }],
            ["interface", { interface: true, implementation: false }],
            ["implementation", { interface: false, implementation: true }]
        ]);
        _this._justTypesOption = new RendererOptions_1.BooleanOption("just-types", "Plain types only", false);
        _this._classPrefixOption = new RendererOptions_1.StringOption("class-prefix", "Class prefix", "PREFIX", "QT");
        _this._extraCommentsOption = new RendererOptions_1.BooleanOption("extra-comments", "Extra comments", false);
        _this.setOptions([
            _this._justTypesOption,
            _this._classPrefixOption,
            _this._featuresOption,
            _this._extraCommentsOption
        ]);
        return _this;
    }
    Object.defineProperty(ObjectiveCTargetLanguage.prototype, "rendererClass", {
        get: function () {
            return ObjectiveCRenderer;
        },
        enumerable: true,
        configurable: true
    });
    return ObjectiveCTargetLanguage;
}(TargetLanguage_1.TargetLanguage));
exports.default = ObjectiveCTargetLanguage;
function typeNameStyle(prefix, original) {
    var words = Strings_1.splitIntoWords(original);
    var result = Strings_1.combineWords(words, legalizeName, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, Strings_1.allUpperWordStyle, Strings_1.allUpperWordStyle, "", isStartCharacter);
    // Take care not to doubly-prefix type names
    return lodash_1.startsWith(result, prefix) ? result : prefix + result;
}
function propertyNameStyle(original) {
    var words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, legalizeName, Strings_1.allLowerWordStyle, Strings_1.firstUpperWordStyle, Strings_1.allLowerWordStyle, Strings_1.allUpperWordStyle, "", isStartCharacter);
}
var keywords = [
    "_Bool",
    "_Complex",
    "_Imaginary",
    "atomic",
    "auto",
    "break",
    "case ",
    "char",
    "const",
    "continue",
    "default",
    "do",
    "double",
    "else",
    "enum",
    "extern",
    "false",
    "float ",
    "for ",
    "goto",
    "if",
    "inline",
    "int",
    "long",
    "nil",
    "nonatomic",
    "register",
    "restrict",
    "retain",
    "return",
    "short",
    "signed",
    "sizeof",
    "static",
    "struct",
    "switch",
    "typedef",
    "true",
    "union",
    "unsigned",
    "void",
    "volatile",
    "while"
];
var propertySafeKeywords = [
    "BOOL",
    "Class",
    "bycopy",
    "byref",
    "decimal",
    "id",
    "IMP",
    "in",
    "inout",
    "NO",
    "NULL",
    "oneway",
    "out",
    "Protocol",
    "SEL",
    "self",
    "super",
    "YES"
];
function isStartCharacter(utf16Unit) {
    return unicode.isAlphabetic(utf16Unit) || utf16Unit === 0x5f; // underscore
}
function isPartCharacter(utf16Unit) {
    var category = unicode.getCategory(utf16Unit);
    return lodash_1.includes(["Nd", "Pc", "Mn", "Mc"], category) || isStartCharacter(utf16Unit);
}
var legalizeName = Strings_1.utf16LegalizeCharacters(isPartCharacter);
function isAnyOrNull(t) {
    return t.kind === "any" || t.kind === "null";
}
var ObjectiveCRenderer = /** @class */ (function (_super) {
    __extends(ObjectiveCRenderer, _super);
    function ObjectiveCRenderer(graph, leadingComments, _justTypes, _classPrefix, _features, _extraComments) {
        var _this = _super.call(this, graph, leadingComments) || this;
        _this._justTypes = _justTypes;
        _this._classPrefix = _classPrefix;
        _this._features = _features;
        _this._extraComments = _extraComments;
        _this.emitBlock = function (line, f) {
            _this.emitLine(line, " {");
            _this.indent(f);
            _this.emitLine("}");
        };
        _this.emitExtraComments = function () {
            var comments = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                comments[_i] = arguments[_i];
            }
            if (!_this._extraComments)
                return;
            for (var _a = 0, comments_1 = comments; _a < comments_1.length; _a++) {
                var comment = comments_1[_a];
                _this.emitLine("// ", comment);
            }
        };
        _this.objcType = function (t) {
            return Type_1.matchType(t, function (_anyType) { return ["id", ""]; }, 
            // For now, we're treating nulls just like any
            function (_nullType) { return ["id", ""]; }, function (_boolType) { return ["NSBoolean", " *"]; }, function (_integerType) { return ["NSNumber", " *"]; }, function (_doubleType) { return ["NSNumber", " *"]; }, function (_stringType) { return ["NSString", " *"]; }, function (arrayType) {
                var itemType = arrayType.items;
                var itemTypeName = _this.objcType(itemType);
                // NSArray<id>* === NSArray*
                if (isAnyOrNull(itemType)) {
                    return ["NSArray", " *"];
                }
                return [["NSArray<", itemTypeName, ">"], " *"];
            }, function (classType) { return [_this.nameForNamedType(classType), " *"]; }, function (mapType) { return [["NSDictionary<NSString *, ", _this.objcType(mapType.values), ">"], " *"]; }, function (enumType) { return [_this.nameForNamedType(enumType), " *"]; }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                return nullable !== null ? _this.objcType(nullable) : ["id", ""];
            });
        };
        _this.jsonType = function (t) {
            return Type_1.matchType(t, function (_anyType) { return ["id", ""]; }, 
            // For now, we're treating nulls just like any
            function (_nullType) { return ["id", ""]; }, function (_boolType) { return ["NSBoolean", " *"]; }, function (_integerType) { return ["NSNumber", " *"]; }, function (_doubleType) { return ["NSNumber", " *"]; }, function (_stringType) { return ["NSString", " *"]; }, function (_arrayType) { return ["NSArray", " *"]; }, function (_classType) { return ["NSDictionary<NSString *, id>", " *"]; }, function (mapType) { return [["NSDictionary<NSString *, ", _this.jsonType(mapType.values), ">"], " *"]; }, function (_enumType) { return ["NSString", " *"]; }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                return nullable !== null ? _this.jsonType(nullable) : ["id", ""];
            });
        };
        _this.fromDynamicExpression = function (t) {
            var dynamic = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                dynamic[_i - 1] = arguments[_i];
            }
            return Type_1.matchType(t, function (_anyType) { return dynamic; }, function (_nullType) { return dynamic; }, function (_boolType) { return dynamic; }, function (_integerType) { return dynamic; }, function (_doubleType) { return dynamic; }, function (_stringType) { return dynamic; }, function (arrayType) { return ["[", dynamic, " map:λ(id x, ", _this.fromDynamicExpression(arrayType.items, "x"), ")]"]; }, function (classType) { return ["[", _this.nameForNamedType(classType), " fromJSONDictionary:", dynamic, "]"]; }, function (mapType) { return ["[", dynamic, " map:λ(id x, ", _this.fromDynamicExpression(mapType.values, "x"), ")]"]; }, function (enumType) { return ["NOT_NIL([", _this.nameForNamedType(enumType), " withValue:", dynamic, "])"]; }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                return nullable !== null ? _this.fromDynamicExpression(nullable, dynamic) : dynamic;
            });
        };
        _this.toDynamicExpression = function (t, dynamic) {
            return Type_1.matchType(t, function (_anyType) { return ["NSNullify(", dynamic, ")"]; }, function (_nullType) { return ["NSNullify(", dynamic, ")"]; }, function (_boolType) { return dynamic; }, function (_integerType) { return dynamic; }, function (_doubleType) { return dynamic; }, function (_stringType) { return dynamic; }, function (arrayType) {
                if (_this.isJSONSafe(arrayType)) {
                    // TODO check each value type
                    return dynamic;
                }
                return ["[", dynamic, " map:λ(id x, ", _this.toDynamicExpression(arrayType.items, "x"), ")]"];
            }, function (_classType) { return ["[", dynamic, " JSONDictionary]"]; }, function (mapType) {
                if (_this.isJSONSafe(mapType)) {
                    // TODO check each value type
                    return dynamic;
                }
                return ["[", dynamic, " map:λ(id x, ", _this.toDynamicExpression(mapType.values, "x"), ")]"];
            }, function (_enumType) { return ["[", dynamic, " value]"]; }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                if (nullable !== null) {
                    if (_this.isJSONSafe(nullable)) {
                        return ["NSNullify(", dynamic, ")"];
                    }
                    else {
                        return ["NSNullify(", _this.toDynamicExpression(nullable, dynamic), ")"];
                    }
                }
                else {
                    // TODO support unions
                    return dynamic;
                }
            });
        };
        _this.safePropertyName = function (propertyName) {
            return Source_1.modifySource(function (serialized) {
                if (lodash_1.includes(propertySafeKeywords, serialized)) {
                    return "self." + serialized;
                }
                else {
                    return serialized;
                }
            }, propertyName);
        };
        _this.emitPropertyAssignment = function (propertyName, _json, propertyType) {
            var key = Strings_1.stringEscape(_json);
            var name = _this.safePropertyName(propertyName);
            Type_1.matchType(propertyType, function (anyType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(anyType, '[dict objectForKey:@"', key, '"]'), ";");
            }, function (nullType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(nullType, '[dict objectForKey:@"', key, '" withClass:[NSNull class]]'), ";");
            }, function (boolType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(boolType, '[dict boolForKey:@"', key, '"]'), ";");
            }, function (integerType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(integerType, '[dict objectForKey:@"', key, '" withClass:[NSNumber class]]'), ";");
            }, function (doubleType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(doubleType, '[dict objectForKey:@"', key, '" withClass:[NSNumber class]]'), ";");
            }, function (stringType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(stringType, '[dict objectForKey:@"', key, '" withClass:[NSString class]]'), ";");
            }, function (arrayType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(arrayType, '[dict objectForKey:@"', key, '" withClass:[NSArray class]]'), ";");
            }, function (classType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(classType, '[dict objectForKey:@"', key, '" withClass:[NSDictionary class]]'), ";");
            }, function (mapType) {
                var itemType = mapType.values;
                if (itemType.isPrimitive() ||
                    // Before union support, we have a lot of untyped data
                    // This ensures that we don't map over unknown elements
                    isAnyOrNull(itemType)) {
                    // TODO check each value type
                    _this.emitLine(name, ' = [dict objectForKey:@"', key, '" withClass:[NSDictionary class]];');
                }
                else {
                    _this.emitLine(name, ' = [[dict objectForKey:@"', key, '" withClass:[NSDictionary class]] map:', ["λ(id x, ", _this.fromDynamicExpression(itemType, "x"), ")"], "];");
                }
            }, 
            // TODO Support enums
            function (enumType) {
                return _this.emitLine(name, " = ", _this.fromDynamicExpression(enumType, '[dict objectForKey:@"', key, '" withClass:[NSString class]]'), ";");
            }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                if (nullable !== null) {
                    _this.ensureBlankLine();
                    // We used to have a NSNullOrNil macro to make this clearer, but we can rely on the
                    // the fact that [nil isNotEqualTo:[NSNull null]] is *false* to not use it
                    _this.emitBlock(['if ([[dict objectForKey:@"', key, '"] isNotEqualTo:[NSNull null]])'], function () {
                        _this.emitPropertyAssignment(propertyName, _json, nullable);
                    });
                    _this.ensureBlankLine();
                }
                else {
                    // TODO This is a union, but for now we just leave it dynamic
                    _this.emitLine(name, " = ", _this.fromDynamicExpression(unionType, '[dict objectForKey:@"', key, '"]'), ";");
                }
                return nullable !== null ? _this.objcType(nullable) : "id";
            });
        };
        _this.emitPrivateClassInterface = function (_, name) {
            _this.emitLine("@interface ", name, " (JSONConversion)");
            _this.emitLine("+ (instancetype)fromJSONDictionary:(NSDictionary *)dict;");
            _this.emitLine("- (NSDictionary *)JSONDictionary;");
            _this.emitLine("@end");
        };
        _this.emitClassInterface = function (t, className) {
            var isTopLevel = _this.topLevels.valueSeq().contains(t);
            _this.emitLine("@interface ", className, " : NSObject");
            _this.forEachClassProperty(t, "none", function (name, _json, propertyType) {
                var attributes = ["nonatomic"];
                // TODO offer a 'readonly' option
                // TODO We must add "copy" if it's NSCopy, otherwise "strong"
                if (propertyType.isNullable) {
                    attributes.push("nullable");
                }
                _this.emitLine("@property ", ["(", attributes.join(", "), ")"], " ", _this.pointerAwareTypeName(propertyType), name, ";");
            });
            if (!_this._justTypes && isTopLevel) {
                if (t.properties.count() > 0)
                    _this.ensureBlankLine();
                _this.emitLine("+ (_Nullable instancetype)fromJSON:(NSString *)json;");
                _this.emitLine("+ (_Nullable instancetype)fromJSON:(NSString *)json encoding:(NSStringEncoding)encoding error:(NSError *_Nullable *)error;");
                _this.emitLine("+ (_Nullable instancetype)fromData:(NSData *)data error:(NSError *_Nullable *)error;");
                _this.emitLine("- (NSString *_Nullable)toJSON;");
                _this.emitLine("- (NSString *_Nullable)toJSON:(NSStringEncoding)encoding error:(NSError *_Nullable *)error;");
                _this.emitLine("- (NSData *_Nullable)toData:(NSError *_Nullable *)error;");
            }
            _this.emitLine("@end");
        };
        // TODO Implement NSCopying
        _this.emitClassImplementation = function (t, className) {
            var isTopLevel = _this.topLevels.valueSeq().contains(t);
            _this.emitLine("@implementation ", className);
            var propertyNames = [];
            _this.forEachClassProperty(t, "none", function (name) {
                if (propertyNames.length !== 0) {
                    propertyNames.push(", ");
                }
                propertyNames.push(name);
                if (propertyNames.length === 5 * 2 - 1) {
                    _this.emitLine("@synthesize ", propertyNames, ";");
                    propertyNames = [];
                }
            });
            if (propertyNames.length !== 0) {
                _this.emitLine("@synthesize ", propertyNames, ";");
            }
            if (!_this._justTypes) {
                _this.ensureBlankLine();
                if (isTopLevel) {
                    _this.emitBlock("+ (_Nullable instancetype)fromData:(NSData *)data error:(NSError *_Nullable *)error", function () {
                        _this.emitLine("return ", className, "FromData(data, error);");
                    });
                    _this.ensureBlankLine();
                    _this.emitBlock("+ (_Nullable instancetype)fromJSON:(NSString *)json encoding:(NSStringEncoding)encoding error:(NSError *_Nullable *)error;", function () {
                        _this.emitLine("return ", className, "FromJSON(json, encoding, error);");
                    });
                    _this.ensureBlankLine();
                    _this.emitBlock("+ (_Nullable instancetype)fromJSON:(NSString *)json", function () {
                        _this.emitLine("NSError *error;");
                        _this.emitLine("return ", className, "FromJSON(json, NSUTF8StringEncoding, &error);");
                    });
                }
                _this.ensureBlankLine();
                _this.emitBlock("+ (instancetype)fromJSONDictionary:(NSDictionary *)dict", function () {
                    _this.emitLine("return [[", className, " alloc] initWithJSONDictionary:dict];");
                });
                _this.ensureBlankLine();
                _this.emitBlock("- (instancetype)initWithJSONDictionary:(NSDictionary *)dict", function () {
                    _this.emitBlock("if (self = [super init])", function () {
                        _this.forEachClassProperty(t, "none", _this.emitPropertyAssignment);
                    });
                    _this.emitLine("return self;");
                });
                _this.ensureBlankLine();
                _this.emitBlock("- (NSDictionary *)JSONDictionary", function () {
                    _this.emitLine("return @{");
                    _this.indent(function () {
                        _this.forEachClassProperty(t, "none", function (propertyName, jsonKey, propertyType) {
                            var key = Strings_1.stringEscape(jsonKey);
                            var name = _this.safePropertyName(propertyName);
                            _this.emitLine('@"', key, '": ', _this.toDynamicExpression(propertyType, name), ",");
                        });
                    });
                    _this.emitLine("};");
                });
                if (isTopLevel) {
                    _this.ensureBlankLine();
                    _this.emitBlock("- (NSData *_Nullable)toData:(NSError *_Nullable *)error", function () {
                        _this.emitLine("return ", className, "ToData(self, error);");
                    });
                    _this.ensureBlankLine();
                    _this.emitBlock("- (NSString *_Nullable)toJSON:(NSStringEncoding)encoding error:(NSError *_Nullable *)error", function () {
                        _this.emitLine("return ", className, "ToJSON(self, encoding, error);");
                    });
                    _this.ensureBlankLine();
                    _this.emitBlock("- (NSString *_Nullable)toJSON", function () {
                        _this.emitLine("NSError *error;");
                        _this.emitLine("return ", className, "ToJSON(self, NSUTF8StringEncoding, &error);");
                    });
                }
            }
            _this.emitLine("@end");
        };
        _this.emitMark = function (label) {
            _this.ensureBlankLine();
            _this.emitLine("// MARK: " + label);
            _this.ensureBlankLine();
        };
        _this.emitDictionaryAndArrayExtensions = function () {
            _this.emitMultiline("@implementation NSArray (JSONConversion)\n- (NSArray *)map:(id (^)(id element))f {\n    id result = [NSMutableArray arrayWithCapacity:self.count];\n    for (id x in self) [result addObject:f(x)];\n    return result;\n}\n@end\n\n@implementation NSDictionary (JSONConversion)\n- (NSDictionary *)map:(id (^)(id value))f {\n    id result = [NSMutableDictionary dictionaryWithCapacity:self.count];\n    for (id key in self) [result setObject:f([self objectForKey:key]) forKey:key];\n    return result;\n}\n\n- (NSException *)exceptionForKey:(id)key type:(NSString *)type {\n    return [NSException exceptionWithName:@\"TypeException\"\n                                    reason:[NSString stringWithFormat:@\"Expected a %@\", type]\n                                    userInfo:@{ @\"dictionary\":self, @\"key\":key }];\n}\n\n- (id)objectForKey:(NSString *)key withClass:(Class)cls {\n    id value = [self objectForKey:key];\n    if ([value isKindOfClass:cls]) return value;\n    else @throw [self exceptionForKey:key type:NSStringFromClass(cls)];\n}\n\n- (NSBoolean *)boolForKey:(NSString *)key {\n    id value = [self objectForKey:key];\n    if ([value isEqual:@(YES)] || [value isEqual:@(NO)]) return value;\n    else @throw [self exceptionForKey:key type:@\"bool\"];\n}\n@end");
        };
        _this.pseudoEnums = graph
            .allTypesUnordered()
            .filter(function (t) { return t instanceof Type_1.EnumType && _this.enumOccursInArrayOrMap(t); })
            .map(function (t) { return t; });
        return _this;
    }
    ObjectiveCRenderer.prototype.enumOccursInArrayOrMap = function (enumType) {
        function containedBy(t) {
            return ((t instanceof Type_1.ArrayType && t.items.equals(enumType)) ||
                (t instanceof Type_1.MapType && t.values.equals(enumType)) ||
                (t instanceof Type_1.ClassType && t.properties.some(containedBy)) ||
                // TODO support unions
                (t instanceof Type_1.UnionType && false));
        }
        return this.typeGraph.allTypesUnordered().some(containedBy);
    };
    Object.defineProperty(ObjectiveCRenderer.prototype, "forbiddenNamesForGlobalNamespace", {
        get: function () {
            return keywords;
        },
        enumerable: true,
        configurable: true
    });
    ObjectiveCRenderer.prototype.forbiddenForClassProperties = function (_c, _classNamed) {
        return { names: [], namespaces: [this.forbiddenWordsNamespace] };
    };
    ObjectiveCRenderer.prototype.forbiddenForEnumCases = function (_e, _enumNamed) {
        return { names: [], namespaces: [this.forbiddenWordsNamespace] };
    };
    ObjectiveCRenderer.prototype.topLevelNameStyle = function (rawName) {
        return Strings_1.camelCase(rawName);
    };
    ObjectiveCRenderer.prototype.makeNamedTypeNamer = function () {
        var _this = this;
        return Naming_1.funPrefixNamer(function (rawName) { return typeNameStyle(_this._classPrefix, rawName); });
    };
    ObjectiveCRenderer.prototype.makeClassPropertyNamer = function () {
        // TODO why is underscore being removed?
        return new Naming_1.Namer(function (s) { return propertyNameStyle(s); }, ["_", "the", "one", "some", "another"]);
    };
    ObjectiveCRenderer.prototype.makeUnionMemberNamer = function () {
        return null;
    };
    ObjectiveCRenderer.prototype.makeEnumCaseNamer = function () {
        return new Naming_1.Namer(propertyNameStyle, []);
    };
    ObjectiveCRenderer.prototype.namedTypeToNameForTopLevel = function (type) {
        return type;
    };
    ObjectiveCRenderer.prototype.emitMethod = function (declaration, f) {
        this.emitLine(declaration);
        this.emitLine("{");
        this.indent(f);
        this.emitLine("}");
    };
    ObjectiveCRenderer.prototype.isJSONSafe = function (t) {
        if (t instanceof Type_1.ClassType) {
            return false;
        }
        else if (t instanceof Type_1.EnumType) {
            return false;
        }
        else if (t instanceof Type_1.ArrayType) {
            return this.isJSONSafe(t.items);
        }
        else if (t instanceof Type_1.MapType) {
            return this.isJSONSafe(t.values);
        }
        else if (t.isPrimitive()) {
            return true;
        }
        else {
            return false;
        }
    };
    ObjectiveCRenderer.prototype.pointerAwareTypeName = function (t) {
        var name = t instanceof Type_1.Type ? this.objcType(t) : t;
        var isPointer = name[1] !== "";
        return isPointer ? name : [name, " "];
    };
    ObjectiveCRenderer.prototype.emitNonClassTopLevelTypedef = function (t, name) {
        var nonPointerTypeName = this.objcType(t)[0];
        this.emitLine("typedef ", nonPointerTypeName, " ", name, ";");
    };
    ObjectiveCRenderer.prototype.topLevelFromDataPrototype = function (name) {
        return [name, " *", name, "FromData(NSData *data, NSError **error)"];
    };
    ObjectiveCRenderer.prototype.topLevelFromJSONPrototype = function (name) {
        return [name, " *", name, "FromJSON(NSString *json, NSStringEncoding encoding, NSError **error)"];
    };
    ObjectiveCRenderer.prototype.topLevelToDataPrototype = function (name) {
        var parameter = this.variableNameForTopLevel(name);
        return ["NSData *", name, "ToData(", name, " *", parameter, ", NSError **error)"];
    };
    ObjectiveCRenderer.prototype.topLevelToJSONPrototype = function (name) {
        var parameter = this.variableNameForTopLevel(name);
        return ["NSString *", name, "ToJSON(", name, " *", parameter, ", NSStringEncoding encoding, NSError **error)"];
    };
    ObjectiveCRenderer.prototype.emitTopLevelFunctionDeclarations = function (_, name) {
        this.emitExtraComments(name);
        this.emitLine(this.topLevelFromDataPrototype(name), ";");
        this.emitLine(this.topLevelFromJSONPrototype(name), ";");
        this.emitLine(this.topLevelToDataPrototype(name), ";");
        this.emitLine(this.topLevelToJSONPrototype(name), ";");
    };
    ObjectiveCRenderer.prototype.emitTopLevelFunctions = function (t, name) {
        var _this = this;
        var parameter = this.variableNameForTopLevel(name);
        this.ensureBlankLine();
        this.emitBlock(this.topLevelFromDataPrototype(name), function () {
            _this.emitLine(_this.pointerAwareTypeName(_this.jsonType(t)), "json = [NSJSONSerialization JSONObjectWithData:data options:NSJSONReadingAllowFragments error:error];");
            _this.emitLine("return *error ? nil : ", _this.fromDynamicExpression(t, "json"), ";");
        });
        this.ensureBlankLine();
        this.emitBlock(this.topLevelFromJSONPrototype(name), function () {
            _this.emitLine("return ", name, "FromData([json dataUsingEncoding:encoding], error);");
        });
        this.ensureBlankLine();
        this.emitBlock(this.topLevelToDataPrototype(name), function () {
            _this.emitLine(_this.pointerAwareTypeName(_this.jsonType(t)), "json = ", _this.toDynamicExpression(t, parameter), ";");
            _this.emitLine("NSData *data = [NSJSONSerialization dataWithJSONObject:json options:kNilOptions error:error];");
            _this.emitLine("return *error ? nil : data;");
        });
        this.ensureBlankLine();
        this.emitBlock(this.topLevelToJSONPrototype(name), function () {
            _this.emitLine("NSData *data = ", name, "ToData(", parameter, ", error);");
            _this.emitLine("return data ? [[NSString alloc] initWithData:data encoding:encoding] : nil;");
        });
    };
    ObjectiveCRenderer.prototype.variableNameForTopLevel = function (name) {
        var _this = this;
        var camelCaseName = Source_1.modifySource(function (serialized) {
            // 1. remove class prefix
            serialized = serialized.substr(_this._classPrefix.length);
            // 2. camel case
            return Strings_1.camelCase(serialized);
        }, name);
        return camelCaseName;
    };
    ObjectiveCRenderer.prototype.emitPseudoEnumInterface = function (enumType, enumName) {
        var _this = this;
        this.emitLine("@interface ", enumName, " : NSObject");
        this.emitLine("@property (nonatomic, readonly, copy) NSString *value;");
        this.emitLine("+ (instancetype _Nullable)withValue:(NSString *)value;");
        this.forEachEnumCase(enumType, "none", function (name, _) {
            _this.emitLine("+ (", enumName, " *)", name, ";");
        });
        this.emitLine("@end");
    };
    ObjectiveCRenderer.prototype.emitPseudoEnumImplementation = function (enumType, enumName) {
        var _this = this;
        this.emitLine("@implementation ", enumName);
        this.emitLine();
        var instances = Source_1.modifySource(function (s) { return pluralize.plural(_this._classPrefix.toLocaleLowerCase() + s.substring(_this._classPrefix.length)); }, enumName);
        this.emitLine("static NSMutableDictionary<NSString *, ", enumName, " *> *", instances, ";");
        this.emitLine("@synthesize value;");
        this.ensureBlankLine();
        this.forEachEnumCase(enumType, "none", function (name, jsonValue) {
            _this.emitLine("+ (", enumName, " *)", name, " { return ", instances, '[@"', Strings_1.stringEscape(jsonValue), '"]; }');
        });
        this.ensureBlankLine();
        this.emitMethod("+ (void)initialize", function () {
            _this.emitLine("NSArray<NSString *> *values = @[");
            _this.indent(function () {
                _this.forEachEnumCase(enumType, "none", function (_, jsonValue) {
                    _this.emitLine('@"', Strings_1.stringEscape(jsonValue), '",');
                });
            });
            _this.emitLine("];");
            _this.emitLine(instances, " = [NSMutableDictionary dictionaryWithCapacity:values.count];");
            _this.emitLine("for (NSString *value in values) ", instances, "[value] = [[", enumName, " alloc] initWithValue:value];");
        });
        this.ensureBlankLine();
        this.emitLine("+ (instancetype _Nullable)withValue:(NSString *)value { return ", instances, "[value]; }");
        this.ensureBlankLine();
        this.emitMethod("- (instancetype)initWithValue:(NSString *)val", function () {
            _this.emitLine("if (self = [super init]) value = val;");
            _this.emitLine("return self;");
        });
        this.ensureBlankLine();
        this.emitLine("- (NSUInteger)hash { return value.hash; }");
        this.emitLine("@end");
    };
    ObjectiveCRenderer.prototype.emitSourceStructure = function () {
        var _this = this;
        if (this.leadingComments !== undefined) {
            this.emitCommentLines("// ", this.leadingComments);
        }
        else if (!this._justTypes && !this._features.interface) {
            this.emitLine("// Remember to import the companion header like:");
            this.emitLine("// ");
            this.emitLine('//   #import "', this._classPrefix, 'Header.h"');
        }
        else if (!this._justTypes) {
            this.emitCommentLines("// ", ["To parse this JSON:", ""]);
            this.emitLine("//   NSError *error;");
            this.forEachTopLevel("none", function (t, topLevelName) {
                var fromJsonExpression = t instanceof Type_1.ClassType
                    ? ["[", topLevelName, " fromJSON:json encoding:NSUTF8Encoding error:&error]"]
                    : [topLevelName, "FromJSON(json, NSUTF8Encoding, &error);"];
                _this.emitLine("//   ", topLevelName, " *", _this.variableNameForTopLevel(topLevelName), " = ", fromJsonExpression);
            });
        }
        this.ensureBlankLine();
        this.emitLine("#import <Foundation/Foundation.h>");
        this.ensureBlankLine();
        if (this._features.interface) {
            if (this._features.implementation) {
                this.emitMark("Types");
            }
            this.ensureBlankLine();
            this.emitExtraComments("This clarifies which properties are JSON booleans");
            this.emitLine("typedef NSNumber NSBoolean;");
            // Emit @class declarations for top-level array+maps and classes
            this.ensureBlankLine();
            this.forEachNamedType("none", function (_, className) { return _this.emitLine("@class ", className, ";"); }, function (_, enumName) { return _this.emitLine("@class ", enumName, ";"); }, function () { return null; });
            this.ensureBlankLine();
            this.ensureBlankLine();
            this.emitLine("NS_ASSUME_NONNULL_BEGIN");
            this.ensureBlankLine();
            this.forEachEnum("leading-and-interposing", function (t, n) { return _this.emitPseudoEnumInterface(t, n); });
            // Emit interfaces for top-level array+maps and classes
            this.forEachTopLevel("leading-and-interposing", function (t, n) { return _this.emitNonClassTopLevelTypedef(t, n); }, function (t) { return !(t instanceof Type_1.ClassType); });
            if (!this._justTypes) {
                this.emitMark("Top-level marshalling functions");
                this.forEachTopLevel("leading-and-interposing", function (t, n) { return _this.emitTopLevelFunctionDeclarations(t, n); });
            }
            this.forEachNamedType("leading-and-interposing", this.emitClassInterface, function () { return null; }, function () { return null; });
            this.ensureBlankLine();
            this.emitLine("NS_ASSUME_NONNULL_END");
        }
        if (this._features.implementation) {
            if (this._features.interface) {
                this.emitMark("Implementation");
            }
            if (!this._justTypes) {
                this.emitExtraComments("Shorthand for simple blocks");
                this.emitLine("#define \u03BB(decl, expr) (^(decl) { return (expr); })");
                this.emitLine("#define NSNullify(x) ([x isNotEqualTo:[NSNull null]] ? x : [NSNull null])");
                this.emitLine();
                this.emitMultiline("id _Nonnull NOT_NIL(id _Nullable x) {\n    if (nil == x) @throw [NSException exceptionWithName:@\"UnexpectedNil\" reason:nil userInfo:nil];\n    return x;\n}");
                this.ensureBlankLine();
                this.emitLine("NS_ASSUME_NONNULL_BEGIN");
                this.ensureBlankLine();
                this.emitDictionaryAndArrayExtensions();
                // We wouldn't need to emit these private iterfaces if we emitted implementations in forward-order
                // but the code is more readable and explicit if we do this.
                if (this._extraComments) {
                    this.emitMark("Private model interfaces");
                }
                this.forEachNamedType("leading-and-interposing", this.emitPrivateClassInterface, function () { return null; }, function () { return null; });
                this.emitMark("Pseudo-enum implementations");
                this.forEachEnum("leading-and-interposing", function (t, n) { return _this.emitPseudoEnumImplementation(t, n); });
                this.emitMark("JSON serialization implementations");
                this.forEachTopLevel("leading-and-interposing", function (t, n) { return _this.emitTopLevelFunctions(t, n); });
            }
            this.forEachNamedType("leading-and-interposing", this.emitClassImplementation, function () { return null; }, function () { return null; });
            if (!this._justTypes) {
                this.ensureBlankLine();
                this.emitLine("NS_ASSUME_NONNULL_END");
            }
        }
    };
    return ObjectiveCRenderer;
}(ConvenienceRenderer_1.ConvenienceRenderer));
