"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Source_1 = require("./Source");
var Support_1 = require("./Support");
var immutable_1 = require("immutable");
var TargetLanguage = /** @class */ (function () {
    function TargetLanguage(displayName, names, extension) {
        var _this = this;
        this.displayName = displayName;
        this.names = names;
        this.extension = extension;
        this.setOptions = function (options) {
            Support_1.assert(_this._options === undefined, "Target language " + _this.displayName + " sets its options more than once");
            _this._options = options;
        };
    }
    Object.defineProperty(TargetLanguage.prototype, "optionDefinitions", {
        get: function () {
            if (this._options === undefined) {
                return Support_1.panic("Target language " + this.displayName + " did not set its options");
            }
            return this._options.map(function (o) { return o.definition; });
        },
        enumerable: true,
        configurable: true
    });
    TargetLanguage.prototype.makeRenderer = function (graph, leadingComments, rendererOptions) {
        if (this._options === undefined) {
            return Support_1.panic("Target language " + this.displayName + " did not set its options");
        }
        return new ((_a = this.rendererClass).bind.apply(_a, [void 0, graph, leadingComments].concat(this._options.map(function (o) { return o.getValue(rendererOptions); }))))();
        var _a;
    };
    TargetLanguage.prototype.renderGraphAndSerialize = function (graph, alphabetizeProperties, leadingComments, rendererOptions, indentation) {
        if (indentation === undefined) {
            indentation = this.defaultIndentation;
        }
        var renderer = this.makeRenderer(graph, leadingComments, rendererOptions);
        if (renderer.setAlphabetizeProperties !== undefined) {
            renderer.setAlphabetizeProperties(alphabetizeProperties);
        }
        var renderResult = renderer.render();
        return Source_1.serializeRenderResult(renderResult, indentation);
    };
    TargetLanguage.prototype.processHandlebarsTemplate = function (graph, rendererOptions, template) {
        var renderer = this.makeRenderer(graph, undefined, rendererOptions);
        var output = renderer.processHandlebarsTemplate(template);
        return { lines: output.split("\n"), annotations: immutable_1.List() };
    };
    Object.defineProperty(TargetLanguage.prototype, "defaultIndentation", {
        get: function () {
            return "    ";
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TargetLanguage.prototype, "partialStringTypeMapping", {
        get: function () {
            return {};
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TargetLanguage.prototype, "stringTypeMapping", {
        get: function () {
            var partial = this.partialStringTypeMapping;
            /* tslint:disable:strict-boolean-expressions */
            return {
                date: partial.date || "string",
                time: partial.time || "string",
                dateTime: partial.dateTime || "string"
            };
            /* tslint:enable */
        },
        enumerable: true,
        configurable: true
    });
    return TargetLanguage;
}());
exports.TargetLanguage = TargetLanguage;
